# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2026 John Heim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
LiteTalkEmulator
Emulate a LiteTalk hardware synth
"""

import logging
import os
import speechd
import sys
from hardwaresynthemulator import HardwareSynthEmulator

class LiteTalkEmulator(HardwareSynthEmulator):

    ReturnChar = 0x0D
    CommandCode = 0x01
    CancelCommand = 0x18
    InterrogationCommand = 0x3F
    ResetCommand = 0x40
    PunctCommand = 0x42
    IndexCommand = 0x45
    RangeCommand = 0x46
    IdCommand = 0x49
    VoiceCommand = 0x4F
    PitchCommand = 0x50
    RateCommand = 0x53
    VolumeCommand = 0x56

    def __init__(self, serialDevice, debug=0, rate=5, pitch=50, punctuation="n", volume=5, range=5):
        super().__init__(serialDevice, debug)   # call Parent.__init__

        self.rate = rate
        self.pitch = pitch
        self.punctuation = punctuation
        self.volume = volume
        self.range = range
        self .commandMode = None
        self.isCancelled = False

        self.setSpeechRate(f"{rate}")
        self.setSpeechPitch(f"{pitch}")
        # self.setSpeechPunctuation(f"{punctuation}")
        # self.setSpeechVolume(f"{volume}")

    #=== Speech  Functions ===#
    def setSpeechRate(self, strValue):
        """
        Converts the LiteTalk speech rate, a number between 0 and 9
        to a Speech Dispatcher speech rate, a number between -100 and +100,
        and sets the Speech Dispatcher speech rate.
        """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.rate + self.toInteger(strValue, 0)
        else:
            newValue = self.toInteger(strValue, self.rate)
        setValue = self.convertWithinRange(newValue, 0, 9)
        try:
            self.speechClient.set_rate(setValue)
            self.rate = newValue
            logging.info(f"Set speech rate to {newValue} ({setValue}).")
        except Exception as error:
            logging.warning(f"Error setting rate to {newValue} ({setValue}): {error}")

    def setSpeechPitch(self, strValue):
        """
        Converts the  LiteTalk pitch, a number between 0 and 99
        to a Speech Dispatcher pitch, a number between -100 and +100,
        and sets the Speech Dispatcher speech pitch.
        """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.pitch + self.toInteger(strValue, 0)
        else:
            newValue = self.toInteger(strValue, self.pitch)
        setValue = self.convertWithinRange(newValue , 0, 99)
        try:
            self.speechClient.set_pitch(setValue)
            self.pitch = newValue
            logging.info(f"Set speech pitch to {newValue} ({setValue}).")
        except Exception as error:
            logging.warning(f"Error setting pitch to {newValue} ({setValue}): {error}")

    def setSpeechPitchRange(self, strValue):
        """
        Takes a LiteTalk pitch range value, a number between 0 and 9
        converts it to Speech Dispatcher value and sets the range of the speech pitch.
        """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.range + self.toIntegert(strValue, 0)
        else:
            newValue = self.toInteger(strValue, self.range)
        setValue = self.convertWithinRange(newValue , 0, 9)
        try:
            self.speechClient.set_pitch_range(setValue)
            self.range = newValue
            logging.info(f"Set speech pitch range to {newValue} ({setValue}).")
        except Exception as error:
            logging.warning(f"Error setting pitch range to {newValue} ({setValue}): {error}")

    def setSpeechVolume(self, strValue):
        """
        Takes a LiteTalk volume setting, a number between 0 and 9
        and converts it to a Speech Dispatcher volume setting, a number between -100 and 100,
        and sets the  Speech dispatcher volume.
    """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.volume + self.toInnteger(strValue, 0)
        else:
            newValue = self.toInteger(strValue, self.volume)
        setValue = self.convertWithinRange(newValue , 0, 9)
        try:
            self.speechClient.set_volume(setValue)
            self.volume = newValue
            logging.info(f"Set speech volume to {newValue} ({setValue}).")
        except Exception as error:
            logging.warning(f"Error setting volume to {newValue} ({setValue}): {error}")

    def setSpeechPunctuation(self, strValue):
        """
        Map LiteTalk punctuation level to Speech Dispatcher punctuation level and set the value.
        """
        mapping = [speechd.PunctuationMode.ALL, speechd.PunctuationMode.MOST, speechd.PunctuationMode.SOME, speechd.PunctuationMode.NONE]
        key = self.toInteger(strValue, 7)
        if key < 0 or key > 15:
            raise ValueError(f"Invalid LiteTalk punctuation level: {strValue}, ({key}).")
        else:
            setValue = mapping[key % 4]
            try:
                self.speechClient.set_punctuation(setValue)
                self.punctuation = key
                logging.info(f"Set punctuation to {key} ({setValue}).")
            except Exception as error:
                logging.warning(f"Error setting  punctuation to {key} ({setValue}): {error}")

    def setVoiceByID(self, voiceID):
        """
        Map LiteTalk voice ID to Speech Dispatcher voice ID and set voice.
        """
        mapping = [
            "default",
            "MALE1",     # Paul
            "MALE2",   # Harry
            "MALE3",   # Frank
            "FEMALE1",   # Betty
            "FEMALE2",   # Ursula
            "FEMALE3",   # Rita
            "CHILD_MALE",   # Kit
            "CHILD_FEMALE",   # Wendy
        ]

        if voiceID >= 0 and voiceID <= 7:
            voice = mapping[voiceID]
            self.speechClient.set_voice(voice)
            logging.info(f"Set voice to {voice}, ID {voiceID}.")

    def safeSpeak(self):
            text = self.received.decode(errors="ignore")
            self.speak(text)
            logging.debug(f"Said '{text}' ...")
            self.received.clear()

    #=== Emulation Functions ===#
    def parse(self, data):
        for byte in data:
            if byte == self.CancelCommand:
                if not self.isCancelled and self.cancelSpeech():
                    self.response.append(self.ReturnChar)

            elif byte == self.CommandCode:
                self.commandMode = True

            elif self.commandMode:
                # Convert lower case command codes to upper case
                if byte >= 0x61 and byte <= 0x7A:
                    byte -= 0x20

                if byte == self.InterrogationCommand:
                    self.response += b"\x00\x20" + self.DeviceIdString.encode("ascii") + b"\r\x7F"
                    self.commandMode = False

                elif byte == self.IdCommand:
                    self.response .append(self.DeviceIdString.encode("ascii") + b"\r")
                    self.commandMode = False

                elif byte == self.IndexCommand:
                    self.response.append(self.ReturnChar)
                    self.commandMode = False

                elif byte == self.ResetCommand:
                    self.setSpeechVolume(b"5")
                    self.setSpeechRate(b"5")
                    self.setSpeechPitch(b"50")
                    self.received.clear()
                    self.commandMode = False

                elif byte == self.RateCommand:
                    self.setSpeechRate(self.received)
                    self.received.clear()
                    self.commandMode = False

                elif byte == self.PitchCommand:
                    self.setSpeechPitch(self.received)
                    self.received.clear()
                    self.commandMode = False

                elif byte == self.RangeCommand:
                    self.setSpeechPitchRange(self.received)
                    self.received.clear()
                    self.commandMode = False

                elif byte == self.VolumeCommand:
                    self.setSpeechVolume(self.received)
                    self.received.clear()
                    self.commandMode = False

                elif byte == self.PunctCommand:
                    self.setSpeechPunctuation(self.received)
                    self.received.clear()
                    self.commandMode = False

                elif byte == self.VoiceCommand:
                    self.setVoiceByID(self.received[0] - ord("0"))
                    self.received.clear()
                    self.commandMode = False

                elif 0x30 <= byte <= 0x39 or byte == 0x2B or byte == 0x2D:
                    self.received.append(byte)

                else:
                    logging.info(f"Unhandled command, '{chr(byte)}'.")
                    self.received.clear()
                    self.commandMode = False

            # Not in commandMode
            elif self.isPrintable(byte):
                self.received.append(byte)

            if byte == 0x00 or byte == self.ReturnChar or byte == self.CommandCode or len(self.received) >= self.MaxBufferSize:
                self.safeSpeak()

            # Klooge to keep from doing 2 or more cancel commands in a row
            self.isCancelled = (byte == self.CancelCommand)

"""
End  of LiteTalkEmulator class
"""



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2026 John G Heim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
HardwareSynthEmulator class definition
Emulates a hardware speech synthesizer.
"""

import inspect
import logging
import os
import select
import speechd
import sys
import termios
import time
import tty

# termios attribute indices (Linux / POSIX)
IFLAG = 0
OFLAG = 1
CFLAG = 2
LFLAG = 3
ISPEED = 4
OSPEED = 5
CC = 6

class HardwareSynthEmulator:

    Identity  = "RPItalk"
    Version = "1.2"
    DeviceIdString = f"{Identity} {Version}\r"
    MaxBufferSize = 4096

    def __init__(self, serialDevice, debug=0):
        self.serialDevice = f"/dev/{serialDevice}"

        self.serialPort = None
        self.received = bytearray()
        self.response = bytearray()

        logging.basicConfig(format="%(message)s")
        self.debugLevel = int(os.getenv("DEBUG", f"{debug}"))
        self.setDebugLevel(self.debugLevel)

        message = f"Starting {self.Identity}, version {self.Version}."
        try:
            self.speechClient = speechd.SSIPClient(self.Identity)
            self.speak(message)
        except Exception as error:
            logging.critical(f"Error initializing Speech Dispatcher: {error}")
            sys.exit(1)
        logging.info(message)

    def setDebugLevel (self, debug):
        self.debugLevel = debug
        if self.debugLevel >= 2:
            logLevel = logging.DEBUG
        elif self.debugLevel >= 1:
            logLevel = logging.INFO
        else:
            logLevel = logging.WARNING
        logging.getLogger().setLevel(logLevel)
        logging.info(f"Debug/log  level set to {self.debugLevel}/{logLevel}.")

    def dumpBytes(self, data: bytes, key: str) -> None:
        numBytes = len(data)
        hexBytes = " ".join(f"{b:02x}" for b in data)
        asciiBytes = "".join(chr(b) if 32 <= b <= 126 else "." for b in data)
        print(f"{key} {numBytes} bytes: {hexBytes} |{asciiBytes}|", file=sys.stderr, flush=True)

    def toInteger(self, bytes, dflt=None):
        try:
            return int(bytes)
        except (TypeError, ValueError):
            return dflt

    def isPrintable(self, byte):
        return 0x20 <= byte <= 0x7E

    def convertWithinRange(self, oldValue, r1m, r1x, r2m=-100, r2x=100):
        """
        Takes a value in one range(like  75 to 650 (the acceptable range for speech rate for a  DECtalk)
        and mathematically converts it to  another range, (like -100 to 100, the range of values for Speech Dispatcher).
        """
        newValue = int((oldValue - (r1x - r1m)/2 - r1m) / ((r1x - r1m)/(r2x - r2m)) + (r2x +r2m)/2)
        return max(r2m, min(r2x, newValue))

    #=== Speech  Functions ===#
    def cancelSpeech(self):
        okay = True
        try:
            self.speechClient.cancel()
        except Exception as error:
            okay = False
        self.received.clear()
        logging.debug("Cancelled speech is {okay}.")
        return okay

    def speak(self, text):
        if text.strip():
            logging.debug(f"Speaking: '{text}' ...")
            self.speechClient.speak(text)

    def endSpeech(self):
        self.speechClient.stop()
        self.speechClient.close()

    def testSpeech(self):
        """
        Reads user input and sends it to Speech Dispatcher.
        """
        print("Type text and press Enter to speak.")
        print("Press Ctrl+D or Ctrl+C to exit.")

        try:
            while True:
                try:
                    text = input("> ")
                    data = text.encode('ascii') + b"\r"
                    response = self.parse(data)
                    logging.info(f"Response is '{response}'.")
                except EOFError:
                    print()
                    break

        except KeyboardInterrupt:
            print()

        finally:
            self.endSpeech()
    #=== Host I/O functions ===#
    def openSerialPort2(self):
        while not os.path.exists(self.serialDevice):
            logging.info(f"Waiting for USB device, {self.serialDevice} ...")
            time.sleep(1.0)

        try:
            self.serialPort = os.open(self.serialDevice, os.O_RDWR | os.O_NOCTTY)
            tty.setraw(self.serialPort)
            attrs = termios.tcgetattr(self.serialPort)
            attrs[CC][termios.VMIN] = 1
            attrs[CC][termios.VTIME] = 0
            attrs[IFLAG] &= ~(termios.IXON | termios.IXOFF | termios.IXANY)
            termios.tcsetattr(self.serialPort, termios.TCSANOW, attrs)
        except OSError as e:
            raise RuntimeError(f"Failed to open/configure serial device {self.serialDevice}: {e}")
        logging.info(f"Opened serial port on device, {self.serialDevice}.")

    def openSerialPort(self):
        while not os.path.exists(self.serialDevice):
            logging.info(f"Waiting for USB device, {self.serialDevice} ...")
            time.sleep(1.0)

        try:
            self.serialPort = os.open(self.serialDevice, os.O_RDWR | os.O_NOCTTY)

            # Prevent CR/LF munging, ^C handling, etc.
            tty.setraw(self.serialPort)

            # Fetch  serial port attributes
            attrs = termios.tcgetattr(self.serialPort)
            attrs[CFLAG] |= (termios.CLOCAL | termios.CREAD)

            #  Set baud  rate to 9600
            attrs[ISPEED] = termios.B9600
            attrs[OSPEED] = termios.B9600

            # Make reads block until at least 1 byte arrives.
            attrs[CC][termios.VMIN] = 1
            attrs[CC][termios.VTIME] = 0

            # Disable software flow control.
            attrs[IFLAG] &= ~(termios.IXON | termios.IXOFF | termios.IXANY)
            termios.tcsetattr(self.serialPort, termios.TCSANOW, attrs)
        except OSError as e:
            raise RuntimeError(f"Failed to open/configure serial device {self.serialDevice}: {e}")
        logging.info(f"Opened serial port on device, {self.serialDevice}.")

    def writeToHost(self, data: bytes) -> bool:
        """
        Write bytes to the host
        Returns True on success, False on failure.
        """
        success = False
        if self.serialPort is None:
            logging.warning("Serial port is None", 1)
        else:
            try:
                bytesRequested = len(data)
                bytesWritten = os.write(self.serialPort, data)
                if bytesWritten != bytesRequested:
                    logging.warn(f"Wrote ({bytesWritten} when {bytesRequested} were to be sent.")
                else:
                    success = True

            except OSError as error:
                logging.warning(f"Write failed: {error}")
        return success

    def testConnection(self):
        while True:
            self.openSerialPort()
            print("Beginning test...", file=sys.stderr, flush=True)
            cnt = 0
            esc = False
            while (self.serialPort):
                try:
                    readableList, _, _ = select.select([self.serialPort], [], [])
                    if self.serialPort in readableList:
                        data = os.read(self.serialPort, 1024)
                        if not data:
                            raise OSError("Host disconnected")
                        else:
                            for byte in data:
                                self.received.append(byte)
                                cnt += 1
                                if byte == 0x5B:
                                    esc = False
                                elif byte == 0x1B:
                                    hexBytes = " ".join(f"{b:02x}" for b in self.received)
                                    asciiBytes = "".join(chr(b) if 32 <= b <= 126 else "." for b in self.received)
                                    print(f"{cnt:02d}: {hexBytes}\n  {esc} '{asciiBytes}'", file=sys.stderr, flush=True)
                                    self.received.clear()
                                    esc = not esc
                                    cnt  = 0

                except OSError:
                    print("Detected host disconnect.", file=sys.stderr, flush=True)
                    os.close(self.serialPort)
                    self.serialPort = None

#=== Emulation functions ===#
    def parse(self, data):

        for byte in data:
            if 0x20 <= byte <= 0x7E:
                self.received.append(byte)

            if byte in [0x01, 0x03, 0x0b, 0x0D] or  len(self.received) >= self.MaxBufferSize:
                self.safeSpeak()

    def emulate(self):
        while True:
            self.openSerialPort()
            logging.info("Beginning emulation...")
            while (self.serialPort):
                try:
                    readableList, _, _ = select.select([self.serialPort], [], [])
                    if self.serialPort in readableList:
                        data = os.read(self.serialPort, 1024)
                        if not data:
                            raise OSError("Host disconnected")
                        else:
                            if self.debugLevel >= 3:
                                self.dumpBytes(data, 'RX')
                            self.parse(data)
                            if self.response:
                                if self.debugLevel >= 3:
                                    self.dumpBytes(self.response, 'AK')
                                self.writeToHost(self.response)
                                self.response.clear()

                except OSError:
                    logging.warn("Detected host disconnect.")
                    os.close(self.serialPort)
                    self.serialPort = None

        self.endSpeech()

"""
End of HardwareSynthEmulator class definition
"""




# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2026 John Heim
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
DECtalkEmulator
Emulate a DECtalk hardware synth
"""

import logging
import os
import re
import speechd
import sys
from hardwaresynthemulator import HardwareSynthEmulator

class DECtalkEmulator(HardwareSynthEmulator):

    FlushChar = 0x01
    BreakChar = 0x03
    IndexChar = 0x0B
    ReturnChar = 0x0D
    XOnChar = 0x11
    XOffChar = 0x13
    CancelChar = 0x18
    EscapeChar = 0x1B
    CommandStart = 0x5B    # left bracket "["
    BackslashChar = 0x5C    # \ backslash
    CommandEnd = 0x5D    # right bracket "]"

    def __init__(self, serialDevice,debug=0, rate=400, pitch=100, punctuation="n", volume=50):
        super().__init__(serialDevice, debug)   # call Parent.__init__

        self.rate = rate
        self.pitch = pitch
        self.punctuation = punctuation
        self.volume = volume
        self.g5 = 0
        self.range =0

        self .commandMode = None
        self.escapeSequence = False
        self.escapePattern = re.compile(r'P\d;\d+;\d+z')

        self.setSpeechRate(f"{rate}")
        self.setSpeechPitch(f"{pitch}")
        self.setSpeechPunctuation(f"{punctuation}")
        self.setSpeechVolume(f"{volume}")

    #===Speech functions ===#
    def setSpeechRate(self, strValue):
        """
        Converts the DECtalk speech rate, a number btween 75 and 650
        to a Speech Dispatcher speech rate, a number between -100 and +100,
        and sets the Speech Dispatcher speech rate.
        """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.rate + self.toInteger(strValue, 0)
        else:
            newValue = self.toInteger(strValue, self.rate)
        setValue = self.convertWithinRange(newValue, 75, 650)
        try:
            self.speechClient.set_rate(setValue)
            self.rate = newValue
            logging.info(f"Set speech rate to {newValue} ({setValue}).")
        except Exception as error:
            logging.warning(f"Error setting rate to {newValue} ({setValue}): {error}")

    def setSpeechPitch(self, strValue):
        """
        Converts the DECtalk speech pitch, a number in mHz btween 50 and 350,
        to a Speech Dispatcher pitch, a number between -100 and +100,
        and sets the Speech Dispatcher speech pitch.
        """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.pitch + self.toInteger(strValue, 0)
        else:
            newValue = self.toInteger(strValue, self.pitch)

        # Klooge for Speakup:
        # Actual range is 50 to 350
        #  but by  setting range to 50 to 180,
        # we put the default, 122,  right in the middle
        # and we make the pitch changes more pronounced.
        setValue = self.convertWithinRange(newValue , 50, 180)
        try:
            self.speechClient.set_pitch(setValue)
            self.pitch = newValue
            logging.info(f"Set speech pitch to {newValue} ({setValue}).")
        except Exception as error:
            logging.warning(f"Error setting pitch to {newValue} ({setValue}): {error}")

    def setSpeechPitchRange(self, strValue):
        """
        Sets the range of the speech pitch.
        The value is a percentage in both DECtalk and Speech Dispatcher so no conversion is necessary.
        """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.range + self.toInteger(strValue, 0)
        else:
            newValue = self.toInteger(strValue, self.range)
        setValue = max(0, min(100, newValue))
        try:
            self.speechClient.set_pitch_range(setValue)
            self.range = newValue
            logging.info(f"Set speech pitch range to {newValue}.")
        except Exception as error:
            logging.warning(f"Error setting pitch range to {newValue}: {error}")

    def setSpeechVolume(self, strValue):
        """
        Converts the DECtalk volume, a number btween 0 and 100,
        to a Speech Dispatcher volume, a number between -100 and +100,
        and sets the Speech Dispatcher speech volume.
        """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.volume + int(strValue)
        else:
            newValue = int(strValue)
        setValue = self.convertWithinRange(newValue, 0, 100)
        try:
            self.speechClient.set_volume(setValue)
            self.volume = newValue
            logging.info(f"Set speech volume to {newValue} ({setValue}).")
        except Exception as error:
            logging.warning(f"Error setting volume to {newValue} ({setValue}): {error}")

    def setSpeechG5(self, strValue):
        """
        Converts the DECtalk design voice g5 value, a number btween 60 and 86,
        to a Speech Dispatcher volume, a number between -100 and +100,
        and sets the Speech Dispatcher speech volume.
        """
        if strValue and (strValue[0] == 0x2B or strValue[0] == 0x2D):
            newValue = self.g5 + int(strValue)
        else:
            newValue = int(strValue)
        setValue = self.convertWithinRange(newValue, 60, 86)
        try:
            self.speechClient.set_volume(setValue)
            self.volume = newValue
            logging.info(f"Set design voice g5 value to {newValue} ({setValue}).")
        except Exception as error:
            logging.warning(f"Error setting g5 value to {newValue} ({setValue}): {error}")

    def setSpeechPunctuation(self, strValue):
        """
        Map DECtalk punctuation level to Speech Dispatcher punctuation level and set the value.
        """
        mapping = {
            "n": speechd.PunctuationMode.NONE,
            "s": speechd.PunctuationMode.SOME,
            "a": speechd.PunctuationMode.ALL,
            "p": speechd.PunctuationMode.NONE,   # closest approximation
        }

        key =strValue.lower()[:1]
        if key not in mapping:
            raise ValueError(f"Invalid DECtalk punctuation level: {strValue}, ({key}).")
        else:
            setValue = mapping[key]
            try:
                self.speechClient.set_punctuation(setValue)
                self.punctuation = key
                logging.info(f"Set punctuation to {key} ({setValue}).")
            except Exception as error:
                logging.warning(f"Error setting  punctuation to {key} ({setValue}): {error}")

    def setVoiceByName(self, voiceName):
        """
        Convert a voice name into a Speech Dispatcher voice number
        and call setVoiceByID
        """
        mapping = {
            "p": 1
        }
        key =voiceName.lower()[:1]
        if key not in mapping:
            raise ValueError(f"Invalid DECtalk voice name: {voiceName}, ({key}).")
        else:
            voiceID = mapping[key]
            self.setVoiceByID(voiceID)

    def setVoiceByID(self, voiceID):
        """
        """
        mapping = [
            "default",
            "MALE1",     # Paul
            "MALE2",   # Harry
            "MALE3",   # Frank
            "MALE3",   # Dennis
            "CHILD_MALE",   # Kit
            "FEMALE1",   # Betty
            "FEMALE2",   # Ursula
            "FEMALE3",   # Rita
            "CHILD_FEMALE",   # Wendy
        ]

        if voiceID >= 0 and voiceID <= 9:
            voice = mapping[voiceID]
            self.speechClient.set_voice(voice)
            logging.info(f"Set voice to {voice}, ID {voiceID}.")

    def safeSpeak(self):
            text = self.received.decode(errors="ignore")
            # Check if an escape sequence snuck into the text
            if self.escapePattern.fullmatch(text):
                self.escapeSequence = True
            else:
                text.strip()
                self.speak(text)
                logging.debug(f"Said '{text}' ...")
            self.received.clear()

    #=== Emulation Functions ===#
    def processCommands(self, buffer):
        logging.debug(f"Processing commands: [{buffer}]")
        commands = buffer.strip().split(":")
        for command in commands:
            if not command:
                continue

            parms = command.split()
            parms += [None] * (3 - len(parms)) # Pad to 3 values
            cmd = parms[0][:2]
            val = parms[1]

            if cmd == "ra" and val != "":
                self.setSpeechRate(val)

            elif cmd == "vo" and val != "":
                self.setSpeechVolume(val)

            elif cmd == "pu" and val != "":
                self.setSpeechPunctuation(val)

            elif cmd == "na" and val != "":
                self.setVoiceByName(val)

            elif cmd == "dv" and val != "":
                sbc = parms[1]
                val = parms[2]

                if sbc == "ap" and val != "":
                        self.setSpeechPitch(val)

                elif sbc == "g5" and val != "":
                    self.setSpeechG5(val)

                elif sbc == "pr" and val != "":
                    self.setSpeechPitchRange(val)

            elif cmd[0] == "n" and cmd[1] >= "0" and cmd[1] <= "9":
                self.setVoiceByID(ord(cmd[1]) - ord("0"))

            else:
                logging.debug(f"Unprocessed command '{command}'/{cmd}.")

    def parse(self, data):

        for byte in data:
            # Flag to indicate when to speak buffered text
            vocalize = False

            # This bloc of code processes the byte.  There is another comment at the end of the bloc of code.
            # Three printable chars are special, \, [, and ]
            if byte == self.BackslashChar:
                pass

            # Left bracket char starts command sequence
            elif byte == self.CommandStart:
                self.commandMode = True
                self.escapeSequence = False
                vocalize = True

            # Right bracket char ends command sequence (although it seems to appear for no reason  occasionally.
            elif byte == self.CommandEnd:
                if self.commandMode:
                    if self.received:
                        self.processCommands(self.received.decode("utf-8"))
                        self.received.clear()
                    self.commandMode = False

            elif self.isPrintable(byte):
                self.received.append(byte)
                if len(self.received) >= self.MaxBufferSize:
                    vocalize = not  self.commandMode

            # Process non-printable characters ...
            elif byte == self.EscapeChar:
                if self.escapeSequence:
                    self.received.clear()
                    self.response.append(self.FlushChar)
                    self.escapeSequence = False
                    print("end escape")
                else:
                    self.escapeSequence = True
                    vocalize = True
                    print("escape")

            elif byte == self.BreakChar or  byte == self.CancelChar:
                if self.cancelSpeech():
                    self.response.append(self.FlushChar)

            elif byte == self.IndexChar:
                self.response.append(self.IndexChar)
                vocalize = True
            # End of byte processing code block.

            if vocalize:
                self.safeSpeak()

"""
End of DECtalkEmulator class
"""
# Uncomment to test
# em = DECtalkEmulator("stdin")
# em.parse(b"Hello world[:xx]")

